//##############################################################################
// volumeicon
//
// config.c - a singleton providing configuration values/functions
// 
// Copyright 2010 Maato
//
// Authors:
//    Maato <maato@softwarebakery.com>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//##############################################################################

#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <assert.h>

#include "config.h"

#define CONFIG_DIRNAME "volumeicon"
#define CONFIG_FILENAME "volumeicon"

static char * m_config_file = NULL;
static char * m_helper_program = NULL;
static gchar * m_theme = NULL;
static gchar * m_channel = NULL;
static int m_stepsize = 0;
static gboolean m_lmb_slider = FALSE;

//##############################################################################
// Set functions
//##############################################################################
void config_set_helper(const gchar * helper)
{
	g_free(m_helper_program);
	m_helper_program = g_strdup(helper);
}

void config_set_theme(const gchar * theme)
{
	g_free(m_theme);
	m_theme = g_strdup(theme);
}

void config_set_channel(const gchar * channel)
{
	g_free(m_channel);
	m_channel = g_strdup(channel);
}

void config_set_stepsize(int stepsize)
{
	m_stepsize = stepsize;
}

void config_set_left_mouse_slider(gboolean active)
{
	m_lmb_slider = active;
}

//##############################################################################
// Get functions
//##############################################################################
const gchar * config_get_helper()
{
	return m_helper_program;
}

const gchar * config_get_theme()
{
	return m_theme;
}

const gchar * config_get_channel()
{
	return m_channel;
}

int config_get_stepsize()
{
	return m_stepsize;
}

gboolean config_get_use_gtk_theme()
{
	return g_strcmp0(m_theme, "Default") == 0 ? TRUE : FALSE;
}

gboolean config_get_left_mouse_slider()
{
	return m_lmb_slider;
}

//##############################################################################
// Static functions
//##############################################################################
static void config_load_default()
{
	if(!m_helper_program)
		config_set_helper("xterm -e 'alsamixer'");
	if(!m_channel)
		config_set_channel(NULL);
	if(!m_stepsize)
		config_set_stepsize(5);
	if(!m_theme)
		config_set_theme("Default");
}

static void config_read()
{
	// Clean up previously loaded configuration values
	m_stepsize = 0;
	g_free(m_helper_program);
	g_free(m_channel);
	g_free(m_theme);

	// Load keys from keyfile
	GKeyFile * kf = g_key_file_new();
	g_key_file_load_from_file(kf, m_config_file, G_KEY_FILE_NONE, NULL);
	m_helper_program = g_key_file_get_value(kf, "StatusIcon", "onclick", NULL);
	m_channel = g_key_file_get_value(kf, "Alsa", "channel", NULL);
	m_stepsize = g_key_file_get_integer(kf, "StatusIcon", "stepsize", NULL);
	m_theme = g_key_file_get_value(kf, "StatusIcon", "theme", NULL);
	m_lmb_slider = g_key_file_get_boolean(kf, "StatusIcon", "lmb_slider", NULL);
	g_key_file_free(kf);

	// Load default values for unset keys
	config_load_default();
}

//##############################################################################
// Setup functions
//##############################################################################
void config_write()
{
	assert(m_config_file != NULL);
	
	GKeyFile * kf = g_key_file_new();
	g_key_file_set_integer(kf, "StatusIcon", "stepsize", m_stepsize);
	g_key_file_set_boolean(kf, "StatusIcon", "lmb_slider", m_lmb_slider);
	if(m_helper_program)
		g_key_file_set_value(kf, "StatusIcon", "onclick", m_helper_program);
	if(m_theme)
		g_key_file_set_value(kf, "StatusIcon", "theme", m_theme);
	if(m_channel)
		g_key_file_set_value(kf, "Alsa", "channel", m_channel);

	gsize length;
	gchar * data = g_key_file_to_data(kf, &length, NULL);
	g_file_set_contents(m_config_file, data, -1, NULL);
	g_free(data);
	g_key_file_free(kf);
}

void config_initialize()
{
	// Build config directory name
	gchar * config_dir = g_build_filename(g_get_user_config_dir(),
		CONFIG_DIRNAME, NULL);
	m_config_file = g_build_filename(config_dir, CONFIG_FILENAME, NULL);

	// Make sure config directory exists
	if(!g_file_test(config_dir, G_FILE_TEST_IS_DIR))
		g_mkdir(config_dir, 0777);

	// If a config file doesn't exist, create one with defaults otherwise
	// read the existing one.
	if(!g_file_test(m_config_file, G_FILE_TEST_EXISTS))
	{
		config_load_default();
		config_write();
	}
	else
	{
		config_read();
	}

	g_free(config_dir);
}

